#pragma once

namespace horizon {
enum class ColorP {
    FROM_LAYER,
    LAYER_HIGHLIGHT,
    LAYER_HIGHLIGHT_LIGHTEN,
    JUNCTION,
    FRAG_ORPHAN,
    AIRWIRE_ROUTER,
    TEXT_OVERLAY,
    HOLE,
    DIMENSION,
    ERROR,
    NET,
    BUS,
    FRAME,
    AIRWIRE,
    PIN,
    PIN_HIDDEN,
    DIFFPAIR,
    SHADOW,
    CONNECTION_LINE,
    NOPOPULATE_X,
    PROJECTION,
    NET_TIE,
    SYMBOL_BOUNDING_BOX,
    N_COLORS,
    // colors after N_COLORS aren't part of the UBO
    BACKGROUND,
    GRID,
    CURSOR_NORMAL,
    CURSOR_TARGET,
    ORIGIN,
    MARKER_BORDER,
    SELECTION_BOX,
    SELECTION_LINE,
    SELECTABLE_OUTER,
    SELECTABLE_INNER,
    SELECTABLE_PRELIGHT,
    SELECTABLE_ALWAYS,
    SEARCH,
    SEARCH_CURRENT
};
}
