.imp-status-menu-button {
  padding: 1px 8px 2px 4px;
  border: 0;
  outline-width: 0;
}

.imp-statusbar {
  border-top: 1px solid @borders;
  padding-left: 10px;
}

.imp-settings-overlay {
  border-radius: 0px;
}

.imp-tool-bar {
  border-radius: 0px;
}

.imp-rule-editor-tiny-button-row {
  min-width: 0px;
  padding: 0px;
}

.imp-rule-editor-tiny-button-column {
  min-height: 0px;
  padding-bottom: 0px;
  padding-top: 0px;
}

.bus-toolbar {
  border-right: 1px solid @borders;
}

.tag-entry-tiny-button {
  min-width: 0px;
  padding: 0px;
}

.imp-search-slider {
  background-color: @theme_base_color;
  border-color: @borders;
  border-radius: 0 0 3px 3px;
  border-width: 0 1px 1px 1px;
  border-style: solid;
}

.imp-selection-filter-tiny-button {
  min-width: 0px;
  min-height: 0px;
  padding: 0px;
}

.app-notification *:link {
  color:#4CC6EF;
}

.horizon-scale-label {
  border-radius:0px;
  border-top-left-radius: 5px;
}

.editor-header-button {
  /*padding: 5px 5px 5px 10px;*/
  padding: 0;
  border: none;
  margin: 0;
  font-weight: bold;
  background: none;
  box-shadow: none;
  text-shadow: none;
  border-radius:0;
}

.editor-header-button > * {
  opacity: 0.5;
}

.editor-header-button:hover > * {
  opacity: 0.75;
}

.editor-header-button:active > * {
  opacity: 1;
}

.editor-header > *:not(:first-child) {
  border-left: 1px solid rgba(0,0,0,.1);
}

.editor-header label:not(.editor-header-spacer) {
  padding: 5px 5px 5px 10px;
}

.editor-header {
  border-bottom: 1px solid rgba(0,0,0,.05);
  background-color: @content_view_bg;
}

.editor-header-spacer {
  padding: 0;
  border: 1px solid rgba(0,0,0,0);
}

.help-button {
  min-height: 0px;
  min-width: 0px;
  padding: 2px;
}

.imp-tool-popover {
  padding-bottom: 0px;
}

.imp-action-bar {
  border-radius:0px;
  border-top-right-radius: 5px;
  border-bottom-right-radius: 5px;
}

.imp-action-bar-expander {
  padding:0px;
  min-width: 0px;
  min-height: 0px;
}

.imp-key-box {
  font-weight: bold;
  border: 2px solid;
  border-radius: 5px;
}

.imp-key-box > label {
 padding-left:.25em;
 padding-right:.25em;
 min-width:1em;
}

.imp-key-box > label:not(:last-child) {
  border-right: 2px solid;
}

.rule-import-header > *:not(:first-child) {
  border-left: 1px solid rgba(0,0,0,.1);
}

.rule-import-header label {
  padding: 5px 10px 5px 10px;
}

.rule-import-header {
  border-bottom: 1px solid rgba(0,0,0,.05);
}

.imp-symbol-expanded-preview {
  border-color: @borders;
  border-radius: 0 0 3px 0px;
  border-width: 0 1px 1px 0px;
  border-style: solid;
  padding: 10px;
}

.imp-property-all-button {
  min-height: 0px;
}

.pool-browser-status-box button {
  min-height:0px;
  border-top:0px;
  border-bottom:0px;
  border-radius:0px;
}

.imp-key-hint-slider {
  background-color: @theme_base_color;
  border-color: @borders;
  border-width: 1px 1px 0px 1px;
  border-style: solid;
  border-radius: 3px 3px 0px 0px;
  padding-top: 3px;
}

.imp-instance-path-box > button, .imp-instance-path-box  button.image-button {box-shadow:none; padding-top:0px; padding-bottom:0px; padding-left: 8px; padding-right:8px; border-radius: 0px; border-top-width: 0px; border-bottom-width: 0px; }
.imp-instance-path-box  button.image-button {padding:1px;}

/* adapted from https://gitlab.gnome.org/GNOME/dconf-editor/-/blob/master/editor/large-pathbar.css */
.imp-instance-path-bar {


  padding:0;


  /* nautilus/src/resources/css/Adwaita.css:71 .path-bar-box uses 200ms */
  transition: background-color 0.3s,
              color            0.3s,
              border-color     0.3s;
  border-color:transparent;
}

.imp-instance-path-bar {
  background-color:@theme_bg_color;
  color           :@theme_fg_color;
  border-color    :@borders;
}

window:backdrop .imp-instance-path-bar {
  background-color:@theme_unfocused_bg_color;
  color           :@theme_unfocused_fg_color;
  border-color    :@unfocused_borders;
}

/*\
* * items generic init and hacks
\*/

/* transition hacks */
box.imp-instance-path-bar         > button,
box.imp-instance-path-bar         > label {
  transition:all 0 ease;
}

/* button config */
.imp-instance-path-bar            > button {
  padding:0 0.1em;
  margin :0 0.1em;

  outline-offset:-4px;

  border-radius:0;

  border:none;
  border-color:transparent;
  box-shadow:none;
  background:none;
  background-color:transparent;
  color:currentColor;
}

/*\
* * border bottom
\*/

.imp-instance-path-bar                > button                                > label {
  border-width:3px 2px;
  border-radius:0;
  padding:2px 4px;
  border-style:solid;
  border-color:transparent;
  transition:border-bottom-color 0.3s,
             padding             0.2s;  /* needs to be a bit faster */
}

.imp-instance-path-bar                > button:hover                          > label {
  border-bottom-color:@borders;
}

.imp-instance-path-bar                > button.active                         > label,
.imp-instance-path-bar                > button.active:hover                   > label {
  border-bottom-color:@theme_selected_bg_color;
}

.imp-instance-path-bar:backdrop       > button.active                         > label,
.imp-instance-path-bar:backdrop       > button.active:hover                   > label {
  border-bottom-color:@borders;
}

/*\
* * special texts "colors"
\*/

/* slash */
.imp-instance-path-bar                > label {
  opacity:0.55;
}
.imp-instance-path-bar:backdrop       > label {
  opacity:0.8;
}

.imp_undo_redo_hint {
  border-radius: 5px;
}


.tnum {
  font-feature-settings: "tnum 1";
}

.clearance-grid spinbutton > button {
  padding-left: 3px;
  padding-right: 3px;
}

.via-def-name {
  border-bottom: 1px solid @borders;
  padding: 5px;
}

.rule-via-definitions .radio {
  background: @content_view_bg;
  border-radius:0px;
  -gtk-outline-radius: 0px;
  border-top:0px;
  border-bottom:0px;
  padding: 0px;
}

.rule-via-definitions .radio:first-child {
  border-left:0px;
}

.rule-via-definitions {
  background: @content_view_bg;
}

.rule-via-definitions .radio:checked  {
  background: @theme_selected_bg_color;
  color:@theme_selected_fg_color;
}

.via-def-header label {
  padding: 5px;
}

.via-def-header label:not(:last-child) {
  border-bottom: 1px solid @borders;
}

.via-def-header {
  background-color: @content_view_bg;
  border-top: 1px solid @borders;
  border-left: 1px solid @borders;
  border-bottom: 1px solid @borders;
}

.via-def-layer {
  padding: 0px 5px;
}

.via-def-layer-box .via-def-layer:not(:last-child) {
  border-bottom: 1px solid @borders;
}
